/*====================================================================
  cfgfile.c -> Encapsulates the config and parse file stuff.

             (c) 2000 Santiago Romero AkA NoP / Compiler
                         sromero@unix-shells.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ====================================================================*/
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <netdb.h>
#include <stdlib.h>
#include <regex.h>

#include "popmail.h"
#include "popsockets.h"
#include "serverwork.h"
#include "cfgfile.h"



/*--------------------------------------------------------------------
   int CreateCFGFile( char *fname );

   This function creates (with user's permission) a basic config file.
--------------------------------------------------------------------*/
int CreateCFGFile( char *fname )
{
   char ans;
   int  numservers, i;
   FILE *fp;
   char server[256], user[256], password[256];
   int  port;

   /* Just ask user and create config file: */
   printf("> Failed to open config file %s\n", fname );
   printf("  This means that you haven't still configured this program or\n");
   printf("  maybe you want to create a new config file (erasing the old one).\n\n");
   printf("  Do you want me to create a basic file for you now?\n");
   printf("  I'll ask you whatever I need to create the file...  (y/n) ");

   ans = GetYesNo();
   if( ans == 0 )
   {
       printf("\n  Ok: read /usr/doc/popmail and create it by yourself.\n\n");
       return(0);
   }

   fp = fopen( fname, "wt" );
   if ( fp == NULL )
      error("Couldn't create config file!\n");

   fprintf( fp, "#-------------------------------------------------------------------\n");
   fprintf( fp, "#  Generic .popmailrc file for your HOME directory.\n");
   fprintf( fp, "#  By Santiago Romero, santiago.romero@iname.com\n");
   fprintf( fp, "#-------------------------------------------------------------------\n");
   fprintf( fp, "\n");

   fprintf( fp, "#--------------------------------------------------------------------\n");
   fprintf( fp, "# Server info. Put here the number of servers and their info.\n");
   fprintf( fp, "#--------------------------------------------------------------------\n");
   fprintf( fp, "# How many accounts are defined in the file\n");

   printf("\n> How many different email accounts do you want to define? (1-n) ");
   scanf("%d", &numservers );
   fgetc(stdin);
   if( numservers <= 0 ) numservers = 1;

   fprintf( fp, "popservers %d\n", numservers);

   for( i=0; i<numservers; i++ )
   {
      fprintf( fp, "\n");
      fprintf(fp, "#----- Account number %d---------------------\n", i+1);

      printf("  Account # %d\n", i+1 );
      printf("   -> Mail server name? (ex: pop.server.com): ");
      fgets( server, 255, stdin );
      printf("   -> User server name? (ex: milogin)       : ");
      fgets( user, 255, stdin );
      printf("   -> Account Password? (ex: secretwrd)     : ");
      fgets( password, 255, stdin );
      printf("   -> POP3 Server port? (usually 110)       : ");
      fflush(stdin);
      scanf( "%d", &port );
      fgetc(stdin);

      fprintf( fp, "server %s\n", server);
      fprintf( fp, "user %s\n", user);
      fprintf( fp, "password %s\n", password);
      fprintf( fp, "port %d\n", port);
   }

   fprintf( fp, "\n");
   fprintf( fp, "#----- And so on... -------------------------\n", i);
   fprintf( fp, "\n");

   fprintf( fp, "#--------------------------------------------------------------------\n");
   fprintf( fp, "# Define here different setup options for popmail\n");
   fprintf( fp, "#--------------------------------------------------------------------\n");
   fprintf( fp, "\n");

   printf("\n> Generic options:\n" );
   printf("   -> Max size a message can be in order to be fetched (in bytes).\n");
   printf("      (Enter 0 to disable this option and fetch all messages) : ");
   scanf("%d", &port );
   fgetc(stdin);

   if ( port > 0 )
       fprintf( fp, "maxsize %d\n", port);
   else
       fprintf( fp, "#maxsize 10000000\n");

   printf("\n   -> Enter now the file name where you will write the filter rules.\n");
   printf("      Enter FULL PATH as /home/youruser/.filterfile or write 0 to\n");
   printf("      disable email filtering: ");
   fgets( server, 255, stdin );
   if( server[0] == '0' && strlen(server) < 3 )
      fprintf( fp, "\n#filterfile /home/youruser/.filterfile\n");
   else
      fprintf( fp, "\nfilterfile %s", server);

   printf("\n   -> Enter now a file name where I will LOG info about deleted msgs.\n");
   printf("      Enter FULL PATH as /home/youruser/.popmail.log or write 0 to\n");
   printf("      disable logging: ");
   fgets( server, 255, stdin );
   if( server[0] == '0' && strlen(server) < 3 )
      fprintf( fp, "\n#logfile /home/youruser/.popmail.log\n\n");
   else
      fprintf( fp, "\nlogfile %s\n", server);

   fclose(fp);
   printf("\n> Now I'm going to exec \"chmod 0600 .popmailrc\" to set secure permissions.\n");
   printf("  This way nobody but you will be able to read your users/passwords: ");
   chmod(fname, S_IRUSR | S_IWUSR | S_IFREG );
   printf("done.\n");

   printf("\n> Done! You can now run popmail. Remember to edit .popmailrc\n");
   printf("   if you want to add email accounts or change any option. :-)\n\n");
}


/*--------------------------------------------------------------------
 int ParseFileForNumServers( FILE * );

 This function searches from the beginning of the file till it
 finds the "popservers num" setup option, and returns it to caller.
--------------------------------------------------------------------*/
int ParseFileForNumServers( FILE *fp )
{
   char done = 0;
   char *where;
   char line[256], numero[40], word[40];
   int numservers;

   /* while we don't find the "popservers number" string, do: */
   while(!done)
   {
      /* get a line */
      if( fgets( line, 255, fp ) != NULL )
      {
         /* if the start of the line is popservers, get the number after it: */
         if( (where=strstr(line, "popservers")) == line )
         {
             if(GetFirstWord( where+strlen("popservers"), 20, word ))
                 numservers = atoi( word );
             else
               error("Error parsing .popmailrc file. Delete it and run popmail.\n");
               done = 1;
         }

         if( feof(fp) )
         {
             error("Error parsing .popmailrc file. Delete it and run popmail.\n");
             done = 1;
         }
      }
   }
   return( numservers );
}



/*--------------------------------------------------------------------
 int ParseFileForServers( FILE *, numservers, servers );

 This function reads all the servers from the config file.
--------------------------------------------------------------------*/
int ParseFileForServers( FILE *fp, int numservers,
                         struct Server **servidores )
{
   char done = 0, anerror = 0;
   char *where;
   char line[256], numero[40], word[128];
   int i, server;

   /* default values of the servers: */
    for( i=0; i< numservers; i++ )
   {
      strcpy(servidores[i]->mailserver, "localhost");
      strcpy(servidores[i]->username,   "root");
      strcpy(servidores[i]->password,   "dontknow");
      servidores[i]->puerto = 110;
      servidores[i]->num_msgs = 0;
   }

   /* Now scan the file for server, username, password and port strings. */
   /* And each time we find a "server" string, increment SERVER variable. */
   server = -1;
   while(!done)
   {
      if( fgets( line, 255, fp ) != NULL )
      {

         if( (where=strstr(line, "server")) == line )
         {
            server++;
            if( GetFirstWord( where+strlen("server"), 127, word ) )
                strcpy(servidores[server]->mailserver, word);
            else
                anerror = 1;
         }

         else if( (where=strstr(line, "user")) == line )
         {
            if( GetFirstWord( where+strlen("user"), 127, word ) )
                strcpy(servidores[server]->username, word);
            else
                anerror = 1;
         }

         else if( (where=strstr(line, "password")) == line )
         {
            if( GetFirstWord( where+strlen("password"), 127, word ) )
                strcpy(servidores[server]->password, word);
            else
                anerror = 1;
         }

         else if( (where=strstr(line, "port")) == line )
         {
            if( server == numservers-1 )
                done = 1;

            if( GetFirstWord( where+strlen("user"), 127, word ) )
                servidores[server]->puerto = atoi(word);
            else
                anerror = 1;
         }

      }

      /* there was any parsing error? */
      if( anerror || feof(fp) )
      {
          error("Error parsing .popmailrc file. Delete it and run popmail.\n");
      }
   }
   return( numservers );
}


/*--------------------------------------------------------------------
 int ParseFileForOptions( FILE *, int *, Filters * );

 This function reads the rest of popmail options, including the
 filter file and rules.
--------------------------------------------------------------------*/
int ParseFileForOptions( FILE *fp, char *logfname, int *maxsize,
                         struct Filters *filters )
{
   char done = 0;
   char *where;
   char line[256], numero[40], word[256];
   int numservers;

   /* by default, filters disabled */
   filters->num_rules = 0;

   /* while we don't find the "popservers number" string, do: */
   while(!done)
   {
      /* get a line */
      if( fgets( line, 255, fp ) != NULL )
      {
         if( line[0] != '#' )
         {
            /* look for the maxsize option */
            if( (where=strstr(line, "maxsize")) == line )
            {
               if(GetFirstWord( where+strlen("maxsize"), 20, word ))
                  *maxsize = atoi( word );
               else
               {
                  done = 1;
                  error("Error parsing .popmailrc file. Delete it and run popmail.\n");
               }
            }

            /* now we try to find the filterfile option */
            else if( (where=strstr(line, "filterfile")) == line )
            {
               if(GetFirstWord( where+strlen("filterfile"), 255, word ))
                   ParseFilterFile( word, filters );
               else
               {
                  done = 1;
                  error("Error parsing .popmailrc file. Delete it and run popmail.\n");
               }
            }

            /* now we try to find the filterfile option */
            else if( (where=strstr(line, "logfile")) == line )
            {
               if( GetFirstWord( where+strlen("logfile"), 255, word ) )
                  strcpy( logfname, word );
               else
               {
                  done = 1;
                  error("Error parsing .popmailrc file. Delete it and run popmail.\n");
               }
            }

         }
      }
      if( feof(fp) )
          done = 1;
   }
   return( numservers );
}


/*--------------------------------------------------------------------
 int ParseFilterFile( char *fname, struct Filters *filters );

 This function reads the filterfile and adds all the filters to the
 Filters structure used by popmail to determine if a message must
 be deleted on the server.
--------------------------------------------------------------------*/
int ParseFilterFile( char *fname, struct Filters *filters )
{
   FILE *fp;
   char *where;
   char done = 0;
   char line[256], word[256];

   /* try to open the filterfile. If fails, disable filters */
   fp = fopen( fname, "rt" );
   if( fp == NULL ) return(0);

   /* success -> read all the filter rules */
   while(!done)
   {
      /* get a line */
      if( fgets( line, 255, fp ) != NULL )
      {

         if( line[0] != '#' )
         {
            /* look for the deny option */
            if( (where=strstr(line, "deny")) == line )
            {
               if(GetFirstWord( where+strlen("deny"), 255, word ))
               {
                  /* add the filter */
                  if( filters->num_rules < MAX_FILTERS )
                  {
                     regcomp( &filters->rules[filters->num_rules].rule, word, 0 );
                     strcpy( filters->rules[filters->num_rules].rule_text, word );

                     /*printf("Regla %d; <%s>\n", filters->num_rules,
                           filters->rules[filters->num_rules].rule_text); */
                     filters->num_rules++;
                  }
                  else
                     printf("\nReached MAX_FILTERS when reading filters...\n");
               }
               else
               {
                  done = 1;
                  error("Error parsing filterfile. Read docs and create it properly.\n");
               }
            }
         }
      }
      if( feof(fp) )  done = 1;
   }
}


/*--------------------------------------------------------------------
 char GetFirstWord( char *buffer, char *word );

 Extracts the first word from a line (ignoring spaces and so on).
--------------------------------------------------------------------*/
char GetFirstWord( char *buffer, int maxsize, char *word )
{
   int i = 0, j = 0;

    while( buffer[i] == ' ' || buffer[i] == '\t' ||
           buffer[i] == '\r' || buffer[i] == '\n' || buffer[i] == '\0' )
    {
        i++;
        if( i > maxsize )   return(0);
    }

   while( buffer[i] != ' ' && buffer[i] != '\t' &&
           buffer[i] != '\r' && buffer[i] != '\n' && buffer[i] != '\0')
   {
        word[j] = buffer[i];
        j++;
        i++;
        if( i > maxsize )  return(0);
   }

   word[j] = '\0';

   return(1);
}


/*--------------------------------------------------------------------
 char GetYesNo( void );

 Waits for YES or NO for english (y/n) and spanish (s/n).
--------------------------------------------------------------------*/
char GetYesNo( void )
{
   char ans;

   fflush(stdin);
   ans = getchar();
   if( ans == 's' || ans == 'S' ||
        ans == 'y' || ans == 'Y' )
           return(1);

   return(0);
}