/*====================================================================
  popmail.c ->   That's my POP3-email-checker-plus-spam-eraser: just
                 a way to browse the contents of your mailboxes and
                 delete emails you don't want to fetch. This program
                 will not fetch the emails (use fetchmail, jfpop or
                 fetchpop to do this), it just applies some rules to
                 the messages IN THE MAILBOX in order to allow you
                 to delete some emails before receiving them (you
                 could, as an example, delete emails with a given
                 FROM: (sender) header so that forget about some
                 common spammers). Take a look at the README to learn
                 about what you can do with POPmail.

  I have rewritten and used some code of mailcheck v0.1 by Steven
  Radack (steve@lconn.net), and JFPOP from *puntero_loco and Ferkys
  (of course, with permission of the original coders :)

             (c) 2000 Santiago Romero AkA NoP / Compiler
                       santiago.romero@iname.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ====================================================================*/
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <regex.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <netdb.h>
#include <stdlib.h>
#include <getopt.h>

#include "popmail.h"
#include "popsockets.h"
#include "serverwork.h"
#include "cfgfile.h"


void ParseArgs(int argc, char **argv, char *mode );

/*---- main function --------------------------------------------------*/
int main(int argc, char *argv[])
{
   int     i, j;
   struct  Server **servidores;
   struct  Filters filters;
   struct  stat buf;
   int     num_servers=1;
   char    *home;
   char    configfile[256], logfname[256];
   FILE    *cfgfile, *logfile;
   char    mode, filtering;
   int     maxsize = -1, key;

   ParseArgs( argc, argv, &mode );
   if( mode != SILENT )
   {
      printf("POPMail - (c) 2000 Santiago Romero -");
      printf(" Published under the GNU License.\n\n");
   }

   /* look for the HOME directory and the config file */
   home = getenv("HOME");
   sprintf(configfile, "%s/.popmailrc", home );

   /* Check file permissions. Should be 0600 (only owner can read and write) */
   if( stat( configfile, &buf ) == 0 )
   {
      if( buf.st_mode != (S_IRUSR | S_IWUSR | S_IFREG) )
      {
         printf("> Your .popmailrc file has wrong permissions. People could read your\n");
         printf("  users & passwords if you don't change it to 0600. Do you want popmail\n");
         printf("  to change it to 0600 for you? (by executing \"chmod 0600\") (y/n) ");
         fflush(stdin);
         key = getchar();
         if( key=='y' || key=='Y' || key=='s' || key=='S' )
         {
            printf("\n> Now I'm going to exec \"chmod 0600 .popmailrc\" to set secure permissions.\n");
            printf("  This way nobody but you will be able to read your users/passwords: ");
            chmod( configfile, S_IRUSR | S_IWUSR | S_IFREG );
            printf("done.\n\n");
         }
         else
         {
            printf("\n> Please change the permissions to 0600 (chmod 0600 .popmailrc) before\n");
            printf("  executing this program, so that people can't read your setup/config file.\n\n");
            return(-4);
         }
      }
   }
   else /* file does not exist? */
   {
       if( errno == ENOENT )
           CreateCFGFile( configfile );
	   else
           printf("\n> stat() returned %d <\n", errno );
       return(-1);
   }

   cfgfile = fopen( configfile, "rt" );
   if( cfgfile == NULL )
   {
       CreateCFGFile( configfile );
       return(-1);
   }
   num_servers = ParseFileForNumServers( cfgfile );

   /*--- Get memory for the server list: ------------------------*/
   servidores = (struct Server **)
           malloc( sizeof(struct Server *)*num_servers);
   if( servidores == NULL )
       error("Error getting memory for the server list.\n");

   for( i=0; i<num_servers; i++ )
   {
       servidores[i] = (struct Server *) malloc( sizeof(struct Server));
       if( servidores[i] == NULL )
       {
          for( j=0; j<i; j++ )
             free(servidores[j]);
          free(servidores);
          error("Error getting memory for a server struct.\n");
       }
   }
   /*------------------------------------------------------------*/

   ParseFileForServers( cfgfile, num_servers, servidores );
   ParseFileForOptions( cfgfile, logfname, &maxsize, &filters );

   logfile = fopen( logfname, "at" );
   if( mode == VERBOSE  )
   {
       if( maxsize > 0 )
          printf(" Max Message Size = %d bytes.\n", maxsize);
       else
          printf(" Filtering messages by size disabled.\n", maxsize);
       if( filters.num_rules > 0 )
          printf(" Filtering capabilites enabled (%d entries).\n", filters.num_rules );
       else
          printf(" Filtering capabilities disabled.\n");
       if( logfile == NULL )
       {
          printf(" Logging capabilites disabled. \n");
       }
       else
          printf(" Logging capabilites enabled. \n");
   }

   fclose( cfgfile );

   /* This is the main function: We log into each server and test
      filter/size rules, showing the results on the screen. */
   for( i=0; i<num_servers; i++ )
      ReadServer( servidores[i], &filters, mode, maxsize, logfile );

   /*--- Free regexp allocated with regcomp() ---------------------*/
   for( i=0; i<filters.num_rules; i++ )
      regfree( &filters.rules[filters.num_rules].rule );

   /*--- Free memory allocated for the server list: -------------*/
   for( i=0; i<num_servers; i++ )
      free(servidores[i]);
   free( servidores );
   /*------------------------------------------------------------*/

   if( logfile != NULL )
      fclose(logfile);

   if( mode != SILENT )
      printf("\n");
}


/*----------------------------------------------------------------
  void ParseArgs(int argc, char **argv, char *mode );
  Analizes the program options using getopt (man 1 getopt)
 ----------------------------------------------------------------*/
void ParseArgs(int argc, char **argv, char *mode )
{
 
   int option_index;
   int c;
   static struct option long_options[6] ={
    {"help", 0, 0, 0},
    {"verbose", 0, 0, 1},
    {"silent", 0, 0, 2},
    {"normal", 0, 0, 3},
    {"rebuildconfig", 0, 0, 4},
    {0, 0, 0, 0}
   };
   char Help[] = {"\
 POPmail - (c) 2000 Santiago Romero - Published under the GNU License.\n\n\
 POP3 email checker and anti-spam filtering. Read /usr/doc/popmail\n\
 and edit your $HOME/.popmailrc file to configure popmail properly.\n\n\
\tUsage:\tpopmail [options]\n\n\
\tOptions:  --silent           Do not report info to the screen.\n\
\t          --normal           Report some info to the screen.\n\
\t          --verbose          Be verbose reporting info.\n\
\t          --rebuildconfig    Creates a new config file.\n\
\t          --help             This help.\n\
"};
   
   extern char *optarg;
   extern int optind, opterr, optopt;
   struct  stat buf;
   char    *home;
   char    configfile[128];

   *mode = NORMAL;
   c = 0;
   while( c != -1)
   {
      option_index = 0;
      c = getopt_long(argc, argv, "h?", long_options, &option_index);
      
      if( c != -1 )
      switch(c)
      {
        case 0:
        case 'h':
        case '?': printf("%s\n", Help);
                 exit(0);

        case 1  :  *mode = VERBOSE; break;
        case 2  :  *mode = SILENT; break;
        case 3  :  *mode = NORMAL; break;
        case 4  :
              printf("POPMail - (c) 2000 Santiago Romero -");
              printf(" Published under the GNU License.\n\n");
              /* look for the HOME directory and the config file */
              home = getenv("HOME");
              sprintf(configfile, "%s/.popmailrc", home );
              if( stat( configfile, &buf ) == 0 )
              {
                  if( remove(configfile) != 0 )
                  {
                     printf("Couldn't remove %s\n", configfile );
                  }
              }
              CreateCFGFile( configfile );
              exit(0);
              break;
      }
   }

}
