/*====================================================================
  serverwork.c -> Encapsulates the main program stuff.

  I have rewritten and used some code of mailcheck v0.1 by Steven
  Radack (steve@lconn.net), and JFPOP from *puntero_loco and Ferkys
  (of course, with permission of the original coders :)

             (c) 2000 Santiago Romero AkA NoP / Compiler
                         sromero@unix-shells.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ====================================================================*/
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <netdb.h>
#include <stdlib.h>
#include <sys/timeb.h>

#include "popmail.h"
#include "popsockets.h"
#include "serverwork.h"


/*--------------------------------------------------------------------
 int ReadServer( struct Server *servidor, mode, maxsize );

 This function receives a correctly (I hope) filled Server struct
 with the pop3 server address, the POP3 port, the username and the
 password to log in and obtains all the info about that server (the
 number and size of messages, their headers, and so on...
 This function returns the number of messages or -1 if it failed.
--------------------------------------------------------------------*/
int ReadServer( struct Server *servidor, struct Filters *filters,
                char mode, int maxsize, FILE *log )
{
   char buffer[4096], token[32], tmp[4096];
   int i, j, k, offs, bytes, done=0, numlines;
   struct timeb pt;

   /* Do login and continue only of returns == 0 */
   if ( POPLogin( servidor, mode ) != 0 )
       ;

   else
   {
      /* USER and PASS accepted -> go for it! */
      if( mode == VERBOSE )
         printf("\n- Connected with POP3 server.\n");

      servidor->num_msgs = 0;
      /* Get the message list (number and size): */
      Sendf( servidor->sockfd, 0, "STAT\r\n");
      GetLineFromSocket( servidor->sockfd, buffer, sizeof(buffer));
      CheckBuffer(buffer);

      strcpy(tmp, buffer);
      GetnWord( buffer, 2, token);
      servidor->num_msgs = atoi(token);
      GetnWord( tmp, 3, token );
      servidor->totalsize = atoi(token);

      if( mode != SILENT )
	       printf("  - The user %s in %s has %d message(s)",
                 servidor->username, servidor->mailserver,
                 servidor->num_msgs);
      if( mode == VERBOSE ) printf(".\n");

      /*--- Get memory for the message list: ----------------------*/
      servidor->msgs = (struct Message **)
               malloc( sizeof(struct Message *)*servidor->num_msgs);
      if( servidor->msgs == NULL )
      {
         error("Error getting memory for the message list.\n");
         POPDisconnect( servidor, mode );
      }

      for( i=0; i<servidor->num_msgs; i++ )
      {
          servidor->msgs[i] = (struct Message *)
              malloc( sizeof(struct Message) );
          if( servidor->msgs[i] == NULL )
          {
              for( j=0; j<i; j++ )
                  free(servidor->msgs[j]);
              free(servidor->msgs);
              error("Error getting memory for a message struct.\n");
              POPDisconnect( servidor, mode );
          }
      }
      /*--------------------------------------------------------------*/

      for( i=0; i<servidor->num_msgs; i++ )
      {

          /* first of all execute a LIST command to get msg size */
          snprintf( tmp, sizeof(tmp), "LIST %d\r\n", i+1);
          Sendf( servidor->sockfd, 0, tmp);
          GetLineFromSocket( servidor->sockfd, buffer, sizeof(buffer));
          CheckBuffer(buffer);
          GetnWord( buffer, 3, token);
          servidor->msgs[i]->nbytes = atoi(token);

          /* now extract it's header and tell user info about the msg */
          GetMessageHeader( servidor, i, mode );
          ExtractMessageHeader( servidor, i );

          /* Display just 40 characters of the subject */
          if( mode == VERBOSE )
          {
             if( strlen(servidor->msgs[i]->subject) < 40 )
             {
                printf("    > %2d: %-50s (%d bytes).\n", i+1,
                       servidor->msgs[i]->subject,
                       servidor->msgs[i]->nbytes );
             }
             else
             {
                strcpy(tmp, servidor->msgs[i]->subject );
                tmp[39] = '\0';
                printf("    > %2d: %-50s (%d bytes).\n", i+1, tmp,
                        servidor->msgs[i]->nbytes );
             }
             printf("          From: %s\n", servidor->msgs[i]->from );
          }
      }

      if( mode == VERBOSE )
         printf("    Total size: ");
      if( mode != SILENT )
	      printf(" (%d bytes).\n", servidor->totalsize );

      /*----> Now comes the filter stuff <----------------------------*/
      for( i=0; i<servidor->num_msgs; i++ )
      {
          /* count message lines */
          numlines = CountLines(servidor->msgs[i]->header);

          /* first check if message is larger than the max size */
          if( servidor->msgs[i]->nbytes > maxsize && maxsize != -1 )
          {
              DeleteMessage(servidor, i, mode );
              if( mode != SILENT )
                  printf("    Message %3d deleted: Size (%d) > MAXSIZE (%d).\n",
                              i+1,  servidor->msgs[i]->nbytes, maxsize );

              /* log if needed */
              if( log!= NULL )
              {
                  ftime(&pt);
                  fprintf(log,
                  "POPMail - Account: %s on %s at %s > Msg %d deleted: %d bytes > MAXSIZE (%d).\n > From: %s\n > Subject: %s.\n\n",
                   servidor->username, servidor->mailserver,
                   ctime(&pt), i+1, servidor->msgs[i]->nbytes, maxsize,
                   servidor->msgs[i]->from, servidor->msgs[i]->subject );
              }

          }
          /* now check the filters themselves */
          else
          {
              offs = 0;
              /* break the header into <numlines> lines and
                 apply all the filters to each line */
              for( k=0; k<numlines-1; k++ )
              {
                 offs = ExtractLine( servidor->msgs[i]->header,
                                        255, offs, tmp );

                 /* apply all the filter rules to each line */
                 for( j=0; j<filters->num_rules; j++ )
	             {

                    if( regexec( &filters->rules[j].rule,
                                 tmp, 0, NULL, 0 ) == 0 )
                    {
                        DeleteMessage( servidor, i, mode );
                        if( mode != SILENT )
                           printf("    Message %3d deleted: Deny %s .\n",
                                   i+1, filters->rules[j].rule_text );

                        /* log if needed */
                        if( log!= NULL )
                        {
                           ftime(&pt);
                           fprintf(log,
                           "POPMail - Account: %s on %s at %s > Msg %d deleted (%d). Deny %s\n > From: %s\n > Subject: %s.\n\n",
                            servidor->username, servidor->mailserver,
                            ctime(&pt), i+1, servidor->msgs[i]->nbytes,
                            filters->rules[j].rule_text,
                            servidor->msgs[i]->from, servidor->msgs[i]->subject );
                        }
                    }
                 }

    	      }

          }
      }
      /*--------------------------------------------------------------*/

      /*--- Free memory allocated for the message list: --------------*/
      for( i=0; i<servidor->num_msgs; i++ )
          free(servidor->msgs[i]);
      free( servidor->msgs );
      /*--------------------------------------------------------------*/

      POPDisconnect( servidor, mode );
      return(servidor->num_msgs);
   }
   return(-1);
}


/*--------------------------------------------------------------------
 int POPConnect( struct Server *servidor )

 This function connects the socket to the POP3 server.
--------------------------------------------------------------------*/
int POPConnect( struct Server *servidor )
{
   struct sockaddr_in addr;
   struct hostent *he = gethostbyname(servidor->mailserver);

   if (!he)
   {
      herror(servidor->mailserver);
      return(-1);
   }

   memset( (char *)&addr, 0, sizeof(struct sockaddr_in) );
   addr.sin_family = AF_INET;
   memcpy( &(addr.sin_addr.s_addr), he->h_addr, he->h_length);
   addr.sin_port=htons(servidor->puerto);
   
   if ( connect(servidor->sockfd, (struct sockaddr *)&addr,
                     sizeof(struct sockaddr) ) < 0)
   {
      Perror("connect()");
      return (-1);
   }
   return(1);
}


/*--------------------------------------------------------------------
 int POPLogin( struct Server *servidor )

 This function logins in the POP3 server. 0=OK
--------------------------------------------------------------------*/
int POPLogin( struct Server *servidor, char mode )
{
   int bytes;
   char buffer[4096], cmd[32];

   /* First we create a socket... */
   if ( (servidor->sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
      perror("socket()");

   /* and then we connect it to the specified server... */
   if (POPConnect(servidor) < 0)
      return(-1);

   bytes = read(servidor->sockfd,buffer,sizeof(buffer));

   /* send username... */
   snprintf(cmd, sizeof(cmd), "USER %s\r\n", servidor->username);
   write( servidor->sockfd, cmd, strlen(cmd) );
   bytes = read( servidor->sockfd, buffer, sizeof(buffer));

   /* send password... */
   snprintf( cmd, sizeof(cmd), "PASS %s\r\n", servidor->password);
   write( servidor->sockfd, cmd, strlen(cmd) );
   bytes = read( servidor->sockfd, buffer, sizeof(buffer));
   buffer[bytes] = 0;

   /* first character of server reply will be + or - */
   if (buffer[0] == '-')
   {
      printf("Password for %s is wrong or mailbox already locked.\n", servidor->username);
      POPDisconnect( servidor, mode );
      close(servidor->sockfd);
      return(-2);
   }

   return(0);
}


/*--------------------------------------------------------------------
 int POPDisconnect( struct Server *servidor )

 This function disconnects the socket from the POP3 server.
 It returns the first character of the server message. It
 can be '+' for OK, and '-' for ERROR.
--------------------------------------------------------------------*/
void POPDisconnect( struct Server *servidor, char mode )
{
   int bytes;
   char buffer[4096];

   /* now we send the QUIT command and wait for +OK */
   Sendf( servidor->sockfd, 0, "QUIT\r\n");
   bytes = read( servidor->sockfd, buffer, sizeof(buffer));
   buffer[bytes] = 0;

   if ( buffer[0] == '+' )
      if( mode==VERBOSE ) printf("- Disconnected from POP3 server.\n");
   else
      if( mode==VERBOSE ) printf("- Error closing connection with POP3 server.\n");

}


/*--------------------------------------------------------------------
 int GetMessageHeader( struct Server *servidor, int num_msg );

 This function receives a server with an already established socket
 and retrieves the <num_msg> message header from it (TOP n 0).
--------------------------------------------------------------------*/
int GetMessageHeader( struct Server *servidor, int num_msg, char mode )
{
   char cmd[128], buffer[1024];
   char header[8192], token[32];
   int bytes, done=0, bytes_read = 0;

   header[0] = buffer[0] = '\0';

   /* send the "TOP 0" command to retrieve the message header */
   snprintf(cmd, sizeof(cmd), "TOP %d 0\r\n", num_msg+1 );
   Sendf( servidor->sockfd, 0, cmd);

   /* Read the "+OK <size> octects" message */
   GetLineFromSocket( servidor->sockfd, buffer, sizeof(buffer));

   if (buffer[0] != '+')
   {
      printf("ERROR: %s\n", buffer );
      POPDisconnect( servidor, mode );
      return(-1);
   }

   /* now read the message header into the msg struct: */
   while( !done )
   {
       GetLineFromSocket( servidor->sockfd, buffer, sizeof(buffer));
       bytes_read += strlen( buffer );
       if( buffer[0] == '.' )
           done = 1;
       else
           strcat( header, buffer );
   }

   /* the following is to make secure the reading :) */
   header[4095] = '0';
   strcpy( servidor->msgs[num_msg]->header, header );
   strcpy( servidor->msgs[num_msg]->header+
           strlen(servidor->msgs[num_msg]->header)-4, "\n.\n\0" );

   servidor->msgs[num_msg]->deleted = 0;
   return(1);
}


/*--------------------------------------------------------------------
 int ExtractMessageHeader( struct Server *servidor, int num_msg );

 This function receives a server with an already filled Msg struct
 and extracts from the header field the From, Subject and Date.
--------------------------------------------------------------------*/
int ExtractMessageHeader( struct Server *servidor, int num_msg )
{
   int posicion=0;
   char linea[2048];

   strcpy(servidor->msgs[num_msg]->from, "none");
   strcpy(servidor->msgs[num_msg]->subject, "none");
   strcpy(servidor->msgs[num_msg]->date, "none");

   do
   {
      /* We now extract the different lines of the header... */
      posicion = ExtractLine( servidor->msgs[num_msg]->header,
                                2047, posicion, linea );
      linea[255] =  '\0';

      /* ... and search for known header fields (From, Date...). */
      if( strstr(linea, "From: ") == linea )
           strcpy(servidor->msgs[num_msg]->from, linea+strlen("From: "));

      else if( strstr(linea, "Subject: ") == linea )
           strcpy(servidor->msgs[num_msg]->subject, linea+strlen("Subject: "));

      else if( strstr(linea, "Date: ") == linea )
           strcpy(servidor->msgs[num_msg]->date, linea+strlen("Date: "));
   }
   while( linea[0] != '.' );

   return(1);    /* All done! */
}



/*--------------------------------------------------------------------
 int DeleteServerMessage( struct Server *servidor, int num_msg );

 Deletes a given message from the server (FULL OPERATION).
--------------------------------------------------------------------*/
int DeleteServerMessage( struct Server *servidor, int num_msg )
{
/*
    char cmd[32], buffer[4096], token[32];
   struct Message *nodo;
   int i, j, bytes;

   printf("\n");

   // Do login and continue only of returns == 0
   if ( POPLogin( servidor ) != 0 )
       ;

   else
    {
        // USER and PASS accepted -> go for it!
        printf("- Connected with POP3 server.\n");

      printf("  - The POP3 mailbox of %s in %s has %d message(s).\n",
                    servidor->username, servidor->mailserver,
                    servidor->num_msgs);
      if( num_msg+1 > servidor->num_msgs )
      {
         printf("  - Message %d does not exist...\n", num_msg+1 );
         POPDisconnect( servidor );
         return(-2);
      }

      // send the "DELE n" command to delete the selected message
      snprintf(cmd, sizeof(cmd), "DELE %d\r\n", num_msg+1 );
       Sendf( servidor->sockfd, 0, cmd);

       // Read the "+OK" message
       GetLineFromSocket( servidor->sockfd, buffer, sizeof(buffer));
      if (buffer[0] != '+')
             return(-1);

      printf("  - Message %d deleted...\n", num_msg+1 );

      POPDisconnect( servidor, mode );
   }
*/
    return(1);
}


/*--------------------------------------------------------------------
 int DeleteMessage( struct Server *servidor, int num_msg );

 Deletes a given message from the server (returns 1=OK, -1=ERR).
--------------------------------------------------------------------*/
int DeleteMessage( struct Server *servidor, int num_msg, char mode )
{
   char cmd[32], buffer[128];
   struct Message *nodo;
   int i;

   /* send the "DELE n" command to delete the selected message */
   snprintf(cmd, sizeof(cmd), "DELE %d\r\n", num_msg+1 );
   Sendf( servidor->sockfd, 0, cmd);

   /* Read the "+OK" message */
   GetLineFromSocket( servidor->sockfd, buffer, sizeof(buffer));
   if (buffer[0] != '+')
   {
       if( mode != SILENT )
           printf("Error: %s\n", buffer );
       return(-1);
   }

   /* Now we eliminate the message from the structure, by copying
      all the pointers to the previous one and dec num_msgs. */
   servidor->msgs[num_msg]->deleted = 1;

   return(1);
}


/*--------------------------------------------------------------------
 void CheckBuffer( char * buffer );

 This function checks the output of an read() command received from
 the POP3 server (firts character will be + or - for right/wrong
 command received (+OK, -ERR).
--------------------------------------------------------------------*/
void CheckBuffer(char *buffer)
{
   if (!((buffer[0] == '+') &&
        (buffer[1] == 'O') &&
        (buffer[2] == 'K')))
   {
      printf("Couldn't init transfer.\n");
      printf("Message from server: %s\n", buffer);
      exit(-1);
   }

}


/*--------------------------------------------------------------------
 int GetnWord(char *s, int n, char *word);

 This function returns the n'th word of the sentence pointed by s.
--------------------------------------------------------------------*/
int GetnWord(char *s, int n, char *word)
{
   /* gets the n'th word of s and puts it into word */
   char **str = &s;
   char *token;

   while (n)
   {
      token = strsep(str," ");
      n--;
   }
   if (token)
   {
      strcpy(word,token);
      return(1);
   }
   else
      return(0);
}


/*--------------------------------------------------------------------
 int ExtractLine(char *buffer, char *line);

 Extracts a line (ended on \n or \r) and returns it to the caller.
 It returns also a pointer (offset) to the next line start (i).
--------------------------------------------------------------------*/
int ExtractLine(char *buffer, int maxlength, int i, char *line)
{
   int j=0;

   while( buffer[i] != '\n' && buffer[i] != '\r' )
   {
      line[j] = buffer[i];
        j++; i++;
   }
   line[j] = '\0';

   while( buffer[i] == '\r' || buffer[i] == '\n' )
         i++;

   return(i);
}


/*--------------------------------------------------------------------
 int CountLines(char *buffer);

 Counts the number of lines (ended on \n or \r) and returns it.
--------------------------------------------------------------------*/
int CountLines( char *buffer )
{
   int j=0, i=0;

   while( buffer[i] != '\0' )
   {
      if( buffer[i] == '\n' ) j++;
      i++;
   }
   return(j);
}

